(function(global, undef) {
	var _refreshTimeoutHandle,
		_options,
		_localStorage = global.localStorage,
		_token = null,
		_tokenIsAnonymous = false,
		_tokenScopes = null;
		_pendingTokenState = null;

	function saveState() {
		_localStorage.setItem('_communityOAuth', JSON.stringify({
			token: _token || {},
			tokenIsAnonymous: _tokenIsAnonymous,
			tokenScopes: _tokenScopes,
			pendingTokenState: _pendingTokenState
		}));
	}

	function loadState() {
		var v = _localStorage.getItem('_communityOAuth');
		if (v == null) {
			_token = null;
			_tokenIsAnonymous = false;
			_tokenScopes = null;
			_pendingTokenState = null;
		} else {
			var state = JSON.parse(v);
			var token = state.token || null;
			if (token && token.expires) {
				token.expires = new Date(token.expires);
				_token = token;
				_tokenScopes = state.tokenScopes;
				_tokenIsAnonymous = !!state.tokenIsAnonymous;
			} else {
				_token = null;
				_tokenScopes = null;
				_tokenIsAnonymous = false;
			}
			_pendingTokenState = state.pendingTokenState;
		}
	}

    async function refreshAccessToken() {
    	globals.clearTimeout(_refreshTimeoutHandle);

    	if (_token != null && _token.expires > new Date() && _token.refreshToken) {
    		var codeVerifierString = codeVerifier();
    		if (codeVerifierString) {
    			var data = new FormData();
    			data.append('grant_type', 'refresh_token');
    			data.append('client_id', _options.clientId);
    			data.append('refresh_token', _token.refreshToken);

				var http = new XMLHttpRequest();
				http.open('POST', _options.communityBaseUrl + 'api.ashx/v2/oauth/token', true);
				http.onreadystatechange = function() {
				    if(http.readyState == 4 && http.status == 200) {
				    	var result = JSON.parse(http.responseText);
				      var expires = new Date();
							expires.setSeconds(expires.getSeconds() + result.expires_in);
							_token = {
								accessToken: result.access_token,
								expires: expires,
								refreshToken: result.refresh_token
							};
							saveState();

							_refreshTimeoutHandle = global.setTimeout(async function() {
								await refreshAccessToken();
							}, Math.max(result.expires_in - 120, 60) * 1000);
				    }
				}
				http.onerror = async function() {
					_options.authorizationError({
						error: 'invalid_request',
						description: null
					});
					await removeAuthorization();
				}
				http.send(data);
			}
    	} else {
    		await removeAuthorization();
    	}
    }

    async function removeAuthorization(loginAnonymously) {
			return new Promise(function(resolve, reject) {
				var cleanupLocalData = function() {
					_token = null;
					_pendingTokenState = null;
					_tokenIsAnonymous = false;
					_tokenScopes = null;
					saveState();
					_options.authorizationChanged();
					resolve();
				}
				if (_token && _token.accessToken) {
					var http = new XMLHttpRequest();
					http.open('DELETE', _options.communityBaseUrl + 'api.ashx/v2/oauth/token', true);
					http.setRequestHeader('Authorization', 'Bearer ' + _token.accessToken);
					http.onreadystatechange = function() {
						if(http.readyState == 4) {
							cleanupLocalData();
						}
					}
					http.onerror = function() {
						_options.authorizationError({
							error: 'invalid_request',
							description: null
						});
						cleanupLocalData();
					}
					http.send();    	
				} else {
					cleanupLocalData();
				}
			});
    }

    async function authorize(anonymous) {
    	// generate the code verifier string for PKCE
			var codeVerifierArray = new Uint32Array(56 / 2);
			global.crypto.getRandomValues(codeVerifierArray);
			var codeVerifierString = Array.from(codeVerifierArray, function (dec) {
				var v = ("0" + dec.toString(16));
				return v.substring(v.length - 3);
			}).join('');

			// hash the code verifier with SHA256 to use as the challenge (and base64url encode it)
			var hashedChallenge = new Uint8Array(await crypto.subtle.digest('SHA-256', new TextEncoder().encode(codeVerifierString)));
			var challengeString = "";
			for (var i = 0; i < hashedChallenge.byteLength; i++) {
				challengeString += String.fromCharCode(hashedChallenge[i]);
			}
			challengeString = btoa(challengeString)
				.replace(/\+/g, "-")
				.replace(/\//g, "_")
				.replace(/=+$/, "");

			// save the code verifier for future validation/token maintenance
			_pendingTokenState = {
				codeVerifier: codeVerifierString,
				scopes: _options.scopes,
				isAnonymous: anonymous
			};
			saveState();

			var scope = '';
			if (_options.scopes != null) {
				scope = '&scope=' + encodeURIComponent(_options.scopes.join(' '));
			}

			// redirect to the authorise endpoint
			global.location = _options.communityBaseUrl + 'api.ashx/v2/oauth/authorize' + (anonymous ? '/anonymous' : '') + '?response_type=code&client_id=' + encodeURIComponent(_options.clientId) + '&redirect_uri=' + encodeURIComponent(_options.url) + '&state=' + encodeURIComponent(_options.state || '') + '&code_challenge=' + encodeURIComponent(challengeString) + '&code_challenge_method=S256' + scope;
    }

	global.communityOAuth = {
		initialize: async function(options) {
			return new Promise(async function(resolve, reject) {
				_options = {
					communityBaseUrl: options.communityBaseUrl || '/',
					clientId: options.clientId || '',
					url: options.url || '/',
					authorizationError: options.authorizationError || function() { },
					authorizationChanged: options.authorizationChanged || function() { },
					enableAnonymous: !!options.enableAnonymous,
					state: options.state || '',
					scopes: (((typeof options.scopes) == 'string') ? options.scopes.split(' ') : options.scopes) || null
				};

				loadState();
				
				// see if this is an OAuth callback

				var query = {};
				var adjustedQuery = '';
				if (global.location.search.length > 0) {
					global.location.search.substr(1).split('#')[0].split('&').forEach(function(keyValue) {
						var splitValues = keyValue.split('=');
						var key = decodeURIComponent(splitValues[0]);
						var value = decodeURIComponent(splitValues[1]);
						query[key] = value;
						if (key != 'code' && key != 'expires_in' && key != 'state' && key != 'error' && key != 'error_description') {
							adjustedQuery += (adjustedQuery.length == 0 ? '?' : '&') + splitValues[0] + '=' + splitValues[1];
						}
					});
				}

				if (query.code && query.expires_in) {
					// it is, clear the existing token and attempt to load a new one

					_token = null;
					_tokenScopes = null;
					saveState();

					if (_pendingTokenState && _pendingTokenState.codeVerifier) {

						var data = new FormData();
						data.append('grant_type', 'authorization_code');
						data.append('code', query.code);
						data.append('redirect_uri', _options.url);
						data.append('client_id', _options.clientId);
						data.append('code_verifier', _pendingTokenState.codeVerifier);

						var http = new XMLHttpRequest();
						http.open('POST', _options.communityBaseUrl + 'api.ashx/v2/oauth/token', true);
						http.onreadystatechange = function() {
							if(http.readyState == 4 && http.status == 200) {
								// token received, save it

								var result = JSON.parse(http.responseText);
								var expires = new Date();
								expires.setSeconds(expires.getSeconds() + result.expires_in);
								_token = {
									accessToken: result.access_token,
									expires: expires,
									refreshToken: result.refresh_token
								};

								_options.state = query.state;
								_tokenIsAnonymous = _pendingTokenState.isAnonymous;
								_tokenScopes = _pendingTokenState.scopes;

								// delete the pendingTokenState
								_pendingTokenState = null;

								saveState();
								_options.authorizationChanged();

								// schedule token refreshing
								_refreshTimeoutHandle = global.setTimeout(async function() {
									await refreshAccessToken();
								}, Math.max(result.expires_in - 120, 60) * 1000);

								// remove the OAuth values from the URL
								global.history.pushState(null, global.document.title, _options.url + adjustedQuery);

								resolve();
							}
						}
						http.onerror = async function() {
							_options.authorizationError({
								error: 'invalid_request',
								description: null
							});
							await removeAuthorization(_options.enableAnonymous);
							reject();
						}
						http.send(data);
					}
				} else if (query.error) {
					_options.authorizationError({
						error: query.error, 
						description: query.error_description
					});

					var wasAnonymousRequest = _pendingTokenState && _pendingTokenState.isAnonymous;

					await removeAuthorization();

					// get anonymous token if enabled
					if (_options.enableAnonymous && !wasAnonymousRequest) {
						await authorize(true);
					} else {
						// remove the OAuth values from the URL
						global.history.pushState(null, global.document.title, _options.url + adjustedQuery);
					}

					resolve();
				} else {
					// check if we still have a valid token

					if (_token != null && _token.expires > new Date()) {
						_refreshTimeoutHandle = global.setTimeout(function() {
							refreshAccessToken();
						}, Math.max(_token.expires.getTime() - ((new Date()).getTime()) - 120, 60) * 1000);
					} else {
						await removeAuthorization();

						// get anonymous token if enabled
						if (_options.enableAnonymous) {
							await authorize(true);
						}
					}

					resolve();
				}
			});
		},
		isAnonymous: function() {
			return _tokenIsAnonymous;
		},
		isAuthorized: function() {
			return _token != null && _token.expires > new Date();
		},
		accessToken: function() {
			if (_token != null && _token.expires > new Date()) {
				return _token.accessToken;
			} else {
				return null;
			}
		},
		expires: function() {
			if (_token != null) {
				return _token.expires;
			} else {
				return null;
			}
		},
		grantedScopes: function() {
			return _tokenScopes;
		},
		unauthorize: async function() {
			await removeAuthorization();

			// get anonymous token if enabled
			if (_options.enableAnonymous) {
				await authorize(true);
			}
		},
		authorize: async function() {
			await removeAuthorization(false);
			await authorize(false);
		},
		state: function(val) {
			if (val === undef) {
				return _options.state;
			} else {
				_options.state = val;
			}
		}
	};
})(window);